from __future__ import division
from collections import OrderedDict
from functools import partial
import gzip
import io
import os
import logging
import os.path

import h5py
import numpy
from picklable_itertools.extras import equizip
from PIL import Image
from scipy.io.matlab import loadmat
from six.moves import zip, xrange
import zmq

from fuel.converters.base import check_exists, progress_bar
from fuel.datasets import H5PYDataset
from fuel.utils.formats import tar_open
from fuel.utils.parallel import producer_consumer
from fuel import config

log = logging.getLogger(__name__)

DEVKIT_ARCHIVE = 'ILSVRC2010_devkit-1.0.tar.gz'
DEVKIT_META_PATH = 'devkit-1.0/data/meta.mat'
DEVKIT_VALID_GROUNDTRUTH_PATH = ('devkit-1.0/data/'
                                 'ILSVRC2010_validation_ground_truth.txt')
PATCH_IMAGES_TAR = 'patch_images.tar'
TEST_GROUNDTRUTH = 'ILSVRC2010_test_ground_truth.txt'
TRAIN_IMAGES_TAR = 'ILSVRC2010_images_train.tar'
VALID_IMAGES_TAR = 'ILSVRC2010_images_val.tar'
TEST_IMAGES_TAR = 'ILSVRC2010_images_test.tar'
IMAGE_TARS = (TRAIN_IMAGES_TAR, VALID_IMAGES_TAR, TEST_IMAGES_TAR,
              PATCH_IMAGES_TAR)
PUBLIC_FILES = TEST_GROUNDTRUTH, DEVKIT_ARCHIVE
ALL_FILES = PUBLIC_FILES + IMAGE_TARS


@check_exists(required_files=ALL_FILES)
def convert_ilsvrc2010(directory, output_directory,
                       output_filename='ilsvrc2010.hdf5',
                       shuffle_seed=config.default_seed):
    """Converter for data from the ILSVRC 2010 competition.

    Source files for this dataset can be obtained by registering at
    [ILSVRC2010WEB].

    Parameters
    ----------
    input_directory : str
        Path from which to read raw data files.
    output_directory : str
        Path to which to save the HDF5 file.
    output_filename : str, optional
        The output filename for the HDF5 file. Default: 'ilsvrc2010.hdf5'.
    shuffle_seed : int or sequence, optional
        Seed for a random number generator used to shuffle the order
        of the training set on disk, so that sequential reads will not
        be ordered by class.

    .. [ILSVRC2010WEB] http://image-net.org/challenges/LSVRC/2010/index

    """
    devkit_path = os.path.join(directory, DEVKIT_ARCHIVE)
    test_groundtruth_path = os.path.join(directory, TEST_GROUNDTRUTH)
    train, valid, test, patch = [os.path.join(directory, fn)
                                 for fn in IMAGE_TARS]
    n_train, valid_groundtruth, test_groundtruth, wnid_map = \
        prepare_metadata(devkit_path, test_groundtruth_path)
    n_valid, n_test = len(valid_groundtruth), len(test_groundtruth)
    output_path = os.path.join(output_directory, output_filename)

    with h5py.File(output_path, 'w') as f:
        log.info('Creating HDF5 datasets...')
        prepare_hdf5_file(f, n_train, n_valid, n_test)
        log.info('Processing training set...')
        process_train_set(f, train, patch, n_train, wnid_map, shuffle_seed)
        log.info('Processing validation set...')
        process_other_set(f, 'valid', valid, patch, valid_groundtruth, n_train)
        log.info('Processing test set...')
        process_other_set(f, 'test', test, patch, test_groundtruth,
                          n_train + n_valid)
        log.info('Done.')

    return (output_path,)


def fill_subparser(subparser):
    """Sets up a subparser to convert the ILSVRC2010 dataset files.

    Parameters
    ----------
    subparser : :class:`argparse.ArgumentParser`
        Subparser handling the `ilsvrc2010` command.

    """
    subparser.add_argument(
        "--shuffle-seed", help="Seed to use for randomizing order of the "
                               "training set on disk.",
        default=config.default_seed, type=int, required=False)
    return convert_ilsvrc2010


def prepare_metadata(devkit_archive, test_groundtruth_path):
    """Extract dataset metadata required for HDF5 file setup.

    Parameters
    ----------
    devkit_archive : str or file-like object
        The filename or file-handle for the gzipped TAR archive
        containing the ILSVRC2010 development kit.
    test_groundtruth_path : str or file-like object
        The filename or file-handle for the text file containing
        the ILSVRC2010 test set ground truth.

    Returns
    -------
    n_train : int
        The number of examples in the training set.
    valid_groundtruth : ndarray, 1-dimensional
        An ndarray containing the validation set groundtruth in terms of
        0-based class indices.
    test_groundtruth : ndarray, 1-dimensional
        An ndarray containing the test groundtruth in terms of 0-based
        class indices.
    wnid_map : dict
        A dictionary that maps WordNet IDs to 0-based class indices.

    """
    # Read what's necessary from the development kit.
    synsets, cost_matrix, raw_valid_groundtruth = read_devkit(devkit_archive)

    # Mapping to take WordNet IDs to our internal 0-999 encoding.
    wnid_map = dict(zip((s.decode('utf8') for s in synsets['WNID']),
                        xrange(1000)))

    # Map the 'ILSVRC2010 ID' to our zero-based ID.
    ilsvrc_id_to_zero_based = dict(zip(synsets['ILSVRC2010_ID'],
                                   xrange(len(synsets))))

    # Map the validation set groundtruth to 0-999 labels.
    valid_groundtruth = [ilsvrc_id_to_zero_based[id_]
                         for id_ in raw_valid_groundtruth]

    # Raw test data groundtruth, ILSVRC2010 IDs.
    raw_test_groundtruth = numpy.loadtxt(test_groundtruth_path,
                                         dtype=numpy.int16)

    # Map the test set groundtruth to 0-999 labels.
    test_groundtruth = [ilsvrc_id_to_zero_based[id_]
                        for id_ in raw_test_groundtruth]

    # Ascertain the number of filenames to prepare appropriate sized
    # arrays.
    n_train = int(synsets['num_train_images'].sum())
    log.info('Training set: {} images'.format(n_train))
    log.info('Validation set: {} images'.format(len(valid_groundtruth)))
    log.info('Test set: {} images'.format(len(test_groundtruth)))
    n_total = n_train + len(valid_groundtruth) + len(test_groundtruth)
    log.info('Total (train/valid/test): {} images'.format(n_total))
    return n_train, valid_groundtruth, test_groundtruth, wnid_map


def create_splits(n_train, n_valid, n_test):
    n_total = n_train + n_valid + n_test
    tuples = {}
    tuples['train'] = (0, n_train)
    tuples['valid'] = (n_train, n_train + n_valid)
    tuples['test'] = (n_train + n_valid, n_total)
    sources = ['encoded_images', 'targets', 'filenames']
    return OrderedDict(
        (split, OrderedDict((source, tuples[split]) for source in sources))
        for split in ('train', 'valid', 'test')
    )


def prepare_hdf5_file(hdf5_file, n_train, n_valid, n_test):
    """Create datasets within a given HDF5 file.

    Parameters
    ----------
    hdf5_file : :class:`h5py.File` instance
        HDF5 file handle to which to write.
    n_train : int
        The number of training set examples.
    n_valid : int
        The number of validation set examples.
    n_test : int
        The number of test set examples.

    """
    n_total = n_train + n_valid + n_test
    splits = create_splits(n_train, n_valid, n_test)
    hdf5_file.attrs['split'] = H5PYDataset.create_split_array(splits)
    vlen_dtype = h5py.special_dtype(vlen=numpy.dtype('uint8'))
    hdf5_file.create_dataset('encoded_images', shape=(n_total,),
                             dtype=vlen_dtype)
    hdf5_file.create_dataset('targets', shape=(n_total, 1), dtype=numpy.int16)
    hdf5_file.create_dataset('filenames', shape=(n_total, 1), dtype='S32')


def process_train_set(hdf5_file, train_archive, patch_archive, n_train,
                      wnid_map, shuffle_seed=None):
    """Process the ILSVRC2010 training set.

    Parameters
    ----------
    hdf5_file : :class:`h5py.File` instance
        HDF5 file handle to which to write. Assumes `features`, `targets`
        and `filenames` already exist and have first dimension larger than
        `n_train`.
    train_archive :  str or file-like object
        Filename or file handle for the TAR archive of training images.
    patch_archive :  str or file-like object
        Filename or file handle for the TAR archive of patch images.
    n_train : int
        The number of items in the training set.
    wnid_map : dict
        A dictionary mapping WordNet IDs to class indices.
    shuffle_seed : int or sequence, optional
        Seed for a NumPy random number generator that permutes the
        training set on disk. If `None`, no permutation is performed
        (this is the default).

    """
    producer = partial(train_set_producer, train_archive=train_archive,
                       patch_archive=patch_archive, wnid_map=wnid_map)
    consumer = partial(image_consumer, hdf5_file=hdf5_file,
                       num_expected=n_train, shuffle_seed=shuffle_seed)
    producer_consumer(producer, consumer)


def _write_to_hdf5(hdf5_file, index, image_filename, image_data,
                   class_index):
    hdf5_file['filenames'][index] = image_filename.encode('ascii')
    hdf5_file['encoded_images'][index] = image_data
    hdf5_file['targets'][index] = class_index


def train_set_producer(socket, train_archive, patch_archive, wnid_map):
    """Load/send images from the training set TAR file or patch images.

    Parameters
    ----------
    socket : :class:`zmq.Socket`
        PUSH socket on which to send loaded images.
    train_archive :  str or file-like object
        Filename or file handle for the TAR archive of training images.
    patch_archive :  str or file-like object
        Filename or file handle for the TAR archive of patch images.
    wnid_map : dict
        A dictionary that maps WordNet IDs to 0-based class indices.
        Used to decode the filenames of the inner TAR files.

    """
    patch_images = extract_patch_images(patch_archive, 'train')
    num_patched = 0
    with tar_open(train_archive) as tar:
        for inner_tar_info in tar:
            with tar_open(tar.extractfile(inner_tar_info.name)) as inner:
                wnid = inner_tar_info.name.split('.')[0]
                class_index = wnid_map[wnid]
                filenames = sorted(info.name for info in inner
                                   if info.isfile())
                images_gen = (load_from_tar_or_patch(inner, filename,
                                                     patch_images)
                              for filename in filenames)
                pathless_filenames = (os.path.split(fn)[-1]
                                      for fn in filenames)
                stream = equizip(pathless_filenames, images_gen)
                for image_fn, (image_data, patched) in stream:
                    if patched:
                        num_patched += 1
                    socket.send_pyobj((image_fn, class_index), zmq.SNDMORE)
                    socket.send(image_data)
    if num_patched != len(patch_images):
        raise ValueError('not all patch images were used')


def image_consumer(socket, hdf5_file, num_expected, shuffle_seed=None,
                   offset=0):
    """Fill an HDF5 file with incoming images from a socket.

    Parameters
    ----------
    socket : :class:`zmq.Socket`
        PULL socket on which to receive images.
    hdf5_file : :class:`h5py.File` instance
        HDF5 file handle to which to write. Assumes `features`, `targets`
        and `filenames` already exist and have first dimension larger than
        `sum(images_per_class)`.
    num_expected : int
        The number of items we expect to be sent over the socket.
    shuffle_seed : int or sequence, optional
        Seed for a NumPy random number generator that permutes the
        images on disk.
    offset : int, optional
        The offset in the HDF5 datasets at which to start writing
        received examples. Defaults to 0.

    """
    with progress_bar('images', maxval=num_expected) as pb:
        if shuffle_seed is None:
            index_gen = iter(xrange(num_expected))
        else:
            rng = numpy.random.RandomState(shuffle_seed)
            index_gen = iter(rng.permutation(num_expected))
        for i, num in enumerate(index_gen):
            image_filename, class_index = socket.recv_pyobj(zmq.SNDMORE)
            image_data = numpy.fromstring(socket.recv(), dtype='uint8')
            _write_to_hdf5(hdf5_file, num + offset, image_filename,
                           image_data, class_index)
            pb.update(i + 1)


def process_other_set(hdf5_file, which_set, image_archive, patch_archive,
                      groundtruth, offset):
    """Process the validation or test set.

    Parameters
    ----------
    hdf5_file : :class:`h5py.File` instance
        HDF5 file handle to which to write. Assumes `features`, `targets`
        and `filenames` already exist and have first dimension larger than
        `sum(images_per_class)`.
    which_set : str
        Which set of images is being processed. One of 'train', 'valid',
        'test'.  Used for extracting the appropriate images from the patch
        archive.
    image_archive : str or file-like object
        The filename or file-handle for the TAR archive containing images.
    patch_archive : str or file-like object
        Filename or file handle for the TAR archive of patch images.
    groundtruth : iterable
        Iterable container containing scalar 0-based class index for each
        image, sorted by filename.
    offset : int
        The offset in the HDF5 datasets at which to start writing.

    """
    producer = partial(other_set_producer, image_archive=image_archive,
                       patch_archive=patch_archive,
                       groundtruth=groundtruth, which_set=which_set)
    consumer = partial(image_consumer, hdf5_file=hdf5_file,
                       num_expected=len(groundtruth), offset=offset)
    producer_consumer(producer, consumer)


def other_set_producer(socket, which_set, image_archive, patch_archive,
                       groundtruth):
    """Push image files read from the valid/test set TAR to a socket.

    Parameters
    ----------
    socket : :class:`zmq.Socket`
        PUSH socket on which to send images.
    which_set : str
        Which set of images is being processed. One of 'train', 'valid',
        'test'.  Used for extracting the appropriate images from the patch
        archive.
    image_archive : str or file-like object
        The filename or file-handle for the TAR archive containing images.
    patch_archive : str or file-like object
        Filename or file handle for the TAR archive of patch images.
    groundtruth : iterable
        Iterable container containing scalar 0-based class index for each
        image, sorted by filename.

    """
    patch_images = extract_patch_images(patch_archive, which_set)
    num_patched = 0
    with tar_open(image_archive) as tar:
        filenames = sorted(info.name for info in tar if info.isfile())
        images = (load_from_tar_or_patch(tar, filename, patch_images)
                  for filename in filenames)
        pathless_filenames = (os.path.split(fn)[-1] for fn in filenames)
        image_iterator = equizip(images, pathless_filenames, groundtruth)
        for (image_data, patched), filename, class_index in image_iterator:
            if patched:
                num_patched += 1
            socket.send_pyobj((filename, class_index), zmq.SNDMORE)
            socket.send(image_data, copy=False)
    if num_patched != len(patch_images):
        raise Exception


def load_from_tar_or_patch(tar, image_filename, patch_images):
    """Do everything necessary to process an image inside a TAR.

    Parameters
    ----------
    tar : `TarFile` instance
        The tar from which to read `image_filename`.
    image_filename : str
        Fully-qualified path inside of `tar` from which to read an
        image file.
    patch_images : dict
        A dictionary containing filenames (without path) of replacements
        to be substituted in place of the version of the same file found
        in `tar`.

    Returns
    -------
    image_data : bytes
        The JPEG bytes representing either the image from the TAR archive
        or its replacement from the patch dictionary.
    patched : bool
        True if the image was retrieved from the patch dictionary. False
        if it was retrieved from the TAR file.

    """
    patched = True
    image_bytes = patch_images.get(os.path.basename(image_filename), None)
    if image_bytes is None:
        patched = False
        try:
            image_bytes = tar.extractfile(image_filename).read()
            numpy.array(Image.open(io.BytesIO(image_bytes)))
        except (IOError, OSError):
            with gzip.GzipFile(fileobj=tar.extractfile(image_filename)) as gz:
                image_bytes = gz.read()
                numpy.array(Image.open(io.BytesIO(image_bytes)))
    return image_bytes, patched


def read_devkit(f):
    """Read relevant information from the development kit archive.

    Parameters
    ----------
    f : str or file-like object
        The filename or file-handle for the gzipped TAR archive
        containing the ILSVRC2010 development kit.

    Returns
    -------
    synsets : ndarray, 1-dimensional, compound dtype
        See :func:`read_metadata_mat_file` for details.
    cost_matrix : ndarray, 2-dimensional, uint8
        See :func:`read_metadata_mat_file` for details.
    raw_valid_groundtruth : ndarray, 1-dimensional, int16
        The labels for the ILSVRC2010 validation set,
        distributed with the development kit code.

    """
    with tar_open(f) as tar:
        # Metadata table containing class hierarchy, textual descriptions, etc.
        meta_mat = tar.extractfile(DEVKIT_META_PATH)
        synsets, cost_matrix = read_metadata_mat_file(meta_mat)

        # Raw validation data groundtruth, ILSVRC2010 IDs. Confusingly
        # distributed inside the development kit archive.
        raw_valid_groundtruth = numpy.loadtxt(tar.extractfile(
            DEVKIT_VALID_GROUNDTRUTH_PATH), dtype=numpy.int16)
    return synsets, cost_matrix, raw_valid_groundtruth


def read_metadata_mat_file(meta_mat):
    """Read ILSVRC2010 metadata from the distributed MAT file.

    Parameters
    ----------
    meta_mat : str or file-like object
        The filename or file-handle for `meta.mat` from the
        ILSVRC2010 development kit.

    Returns
    -------
    synsets : ndarray, 1-dimensional, compound dtype
        A table containing ILSVRC2010 metadata for the "synonym sets"
        or "synsets" that comprise the classes and superclasses,
        including the following fields:
         * `ILSVRC2010_ID`: the integer ID used in the original
           competition data.
         * `WNID`: A string identifier that uniquely identifies
           a synset in ImageNet and WordNet.
         * `wordnet_height`: The length of the longest path to
           a leaf node in the FULL ImageNet/WordNet hierarchy
           (leaf nodes in the FULL ImageNet/WordNet hierarchy
           have `wordnet_height` 0).
         * `gloss`: A string representation of an English
           textual description of the concept represented by
           this synset.
         * `num_children`: The number of children in the hierarchy
           for this synset.
         * `words`: A string representation, comma separated,
           of different synoym words or phrases for the concept
           represented by this synset.
         * `children`: A vector of `ILSVRC2010_ID`s of children
           of this synset, padded with -1. Note that these refer
           to `ILSVRC2010_ID`s from the original data and *not*
           the zero-based index in the table.
         * `num_train_images`: The number of training images for
           this synset.
    cost_matrix : ndarray, 2-dimensional, uint8
        A 1000x1000 matrix containing the precomputed pairwise
        cost (based on distance in the hierarchy) for all
        low-level synsets (i.e. the thousand possible output
        classes with training data associated).

    """
    mat = loadmat(meta_mat, squeeze_me=True)
    synsets = mat['synsets']
    cost_matrix = mat['cost_matrix']
    new_dtype = numpy.dtype([
        ('ILSVRC2010_ID', numpy.int16),
        ('WNID', ('S', max(map(len, synsets['WNID'])))),
        ('wordnet_height', numpy.int8),
        ('gloss', ('S', max(map(len, synsets['gloss'])))),
        ('num_children', numpy.int8),
        ('words', ('S', max(map(len, synsets['words'])))),
        ('children', (numpy.int8, max(synsets['num_children']))),
        ('num_train_images', numpy.uint16)
    ])
    new_synsets = numpy.empty(synsets.shape, dtype=new_dtype)
    for attr in ['ILSVRC2010_ID', 'WNID', 'wordnet_height', 'gloss',
                 'num_children', 'words', 'num_train_images']:
        new_synsets[attr] = synsets[attr]
    children = [numpy.atleast_1d(ch) for ch in synsets['children']]
    padded_children = [
        numpy.concatenate((c,
                           -numpy.ones(new_dtype['children'].shape[0] - len(c),
                                       dtype=numpy.int16)))
        for c in children
    ]
    new_synsets['children'] = padded_children
    return new_synsets, cost_matrix


def extract_patch_images(f, which_set):
    """Extracts a dict of the "patch images" for ILSVRC2010.

    Parameters
    ----------
    f : str or file-like object
        The filename or file-handle to the patch images TAR file.
    which_set : str
        Which set of images to extract. One of 'train', 'valid', 'test'.

    Returns
    -------
    dict
        A dictionary contains a mapping of filenames (without path) to a
        bytes object containing the replacement image.

    Notes
    -----
    Certain images in the distributed archives are blank, or display
    an "image not available" banner. A separate TAR file of
    "patch images" is distributed with the corrected versions of
    these. It is this archive that this function is intended to read.

    """
    if which_set not in ('train', 'valid', 'test'):
        raise ValueError('which_set must be one of train, valid, or test')
    which_set = 'val' if which_set == 'valid' else which_set
    patch_images = {}
    with tar_open(f) as tar:
        for info_obj in tar:
            if not info_obj.name.endswith('.JPEG'):
                continue
            # Pretty sure that '/' is used for tarfile regardless of
            # os.path.sep, but I officially don't care about Windows.
            tokens = info_obj.name.split('/')
            file_which_set = tokens[-2]
            if file_which_set != which_set:
                continue
            filename = tokens[-1]
            patch_images[filename] = tar.extractfile(info_obj.name).read()
    return patch_images



from fuel.downloaders.base import default_downloader


def fill_subparser(subparser):
    """Sets up a subparser to download the CIFAR-100 dataset file.

    The CIFAR-100 dataset file is downloaded from Alex Krizhevsky's
    website [ALEX].

    .. [ALEX] http://www.cs.toronto.edu/~kriz/cifar-100-python.tar.gz

    Parameters
    ----------
    subparser : :class:`argparse.ArgumentParser`
        Subparser handling the `cifar100` command.

    """
    url = 'http://www.cs.toronto.edu/~kriz/cifar-100-python.tar.gz'
    filename = 'cifar-100-python.tar.gz'
    subparser.set_defaults(urls=[url], filenames=[filename])
    return default_downloader



# -*- coding: utf-8 -*-
from fuel.datasets import H5PYDataset
from fuel.utils import find_in_data_path


class BinarizedMNIST(H5PYDataset):
    u"""Binarized, unlabeled MNIST dataset.

    MNIST (Mixed National Institute of Standards and Technology) [LBBH] is
    a database of handwritten digits. It is one of the most famous datasets
    in machine learning and consists of 60,000 training images and 10,000
    testing images. The images are grayscale and 28 x 28 pixels large.

    This particular version of the dataset is the one used in R.
    Salakhutdinov's DBN paper [DBN] as well as the VAE and NADE papers, and
    is accessible through Hugo Larochelle's public website [HUGO].

    The training set has further been split into a training and a
    validation set. All examples were binarized by sampling from a binomial
    distribution defined by the pixel values.

    .. [LBBH] Yann LeCun, Léon Bottou, Yoshua Bengio, and Patrick Haffner,
       *Gradient-based learning applied to document recognition*,
       Proceedings of the IEEE, November 1998, 86(11):2278-2324.

    Parameters
    ----------
    which_sets : tuple of str
        Which split to load. Valid values are 'train', 'valid' and 'test',
        corresponding to the training set (50,000 examples), the validation
        set (10,000 samples) and the test set (10,000 examples).

    """
    filename = 'binarized_mnist.hdf5'

    def __init__(self, which_sets, load_in_memory=True, **kwargs):
        super(BinarizedMNIST, self).__init__(
            file_or_path=find_in_data_path(self.filename),
            which_sets=which_sets,
            load_in_memory=load_in_memory, **kwargs)



# -*- coding: utf-8 -*-
from fuel.utils import find_in_data_path
from fuel.datasets import H5PYDataset


class CalTech101Silhouettes(H5PYDataset):
    u"""CalTech 101 Silhouettes dataset.

    This dataset provides the `split1` train/validation/test split of the
    CalTech101 Silhouette dataset prepared by Benjamin M. Marlin [MARLIN].

    This class provides both the 16x16 and the 28x28 pixel sized version.
    The 16x16 version contains 4082 examples in the training set, 2257
    examples in the validation set and 2302 examples in the test set. The
    28x28 version contains 4100, 2264 and 2307 examples in the train, valid
    and test set.

    Parameters
    ----------
    which_sets : tuple of str
        Which split to load. Valid values are 'train', 'valid' and 'test'.
    size : {16, 28}
        Either 16 or 28 to select the 16x16 or 28x28 pixels version
        of the dataset (default: 28).

    """
    def __init__(self, which_sets, size=28, load_in_memory=True, **kwargs):
        if size not in (16, 28):
            raise ValueError('size must be 16 or 28')

        self.filename = 'caltech101_silhouettes{}.hdf5'.format(size)
        super(CalTech101Silhouettes, self).__init__(
            self.data_path, which_sets=which_sets,
            load_in_memory=load_in_memory, **kwargs)

    @property
    def data_path(self):
        return find_in_data_path(self.filename)



from collections import OrderedDict
from io import BytesIO
import numpy
from numpy.testing import assert_raises, assert_equal
from PIL import Image
from picklable_itertools.extras import partition_all
from six.moves import zip
from fuel import config
from fuel.datasets.base import IndexableDataset
from fuel.schemes import ShuffledScheme, SequentialExampleScheme
from fuel.streams import DataStream
from fuel.transformers.image import (ImagesFromBytes,
                                     MinimumImageDimensions,
                                     RandomFixedSizeCrop,
                                     Random2DRotation)


def reorder_axes(shp):
    if len(shp) == 3:
        shp = (shp[-1],) + shp[:-1]
    elif len(shp) == 2:
        shp = (1,) + shp
    return shp


class ImageTestingMixin(object):
    def common_setup(self):
        ex_scheme = SequentialExampleScheme(self.dataset.num_examples)
        self.example_stream = DataStream(self.dataset,
                                         iteration_scheme=ex_scheme)
        self.batch_size = 2
        scheme = ShuffledScheme(self.dataset.num_examples,
                                batch_size=self.batch_size)
        self.batch_stream = DataStream(self.dataset, iteration_scheme=scheme)


class TestImagesFromBytes(ImageTestingMixin):
    def setUp(self):
        rng = numpy.random.RandomState(config.default_seed)
        self.shapes = [
            (10, 12, 3),
            (9, 8, 4),
            (12, 14, 3),
            (4, 7),
            (9, 8, 4),
            (7, 9, 3)
        ]
        pil1 = Image.fromarray(rng.random_integers(0, 255,
                                                   size=self.shapes[0])
                               .astype('uint8'), mode='RGB')
        pil2 = Image.fromarray(rng.random_integers(0, 255,
                                                   size=self.shapes[1])
                               .astype('uint8'), mode='CMYK')
        pil3 = Image.fromarray(rng.random_integers(0, 255,
                                                   size=self.shapes[2])
                               .astype('uint8'), mode='RGB')
        pil4 = Image.fromarray(rng.random_integers(0, 255,
                                                   size=self.shapes[3])
                               .astype('uint8'), mode='L')
        pil5 = Image.fromarray(rng.random_integers(0, 255,
                                                   size=self.shapes[4])
                               .astype('uint8'), mode='RGBA')
        pil6 = Image.fromarray(rng.random_integers(0, 255,
                                                   size=self.shapes[5])
                               .astype('uint8'), mode='YCbCr')
        source1 = [pil1, pil2, pil3]
        source2 = [pil4, pil5, pil6]
        bytesio1 = [BytesIO() for _ in range(3)]
        bytesio2 = [BytesIO() for _ in range(3)]
        formats1 = ['PNG', 'JPEG', 'BMP']
        formats2 = ['GIF', 'PNG', 'JPEG']
        for s, b, f in zip(source1, bytesio1, formats1):
            s.save(b, format=f)
        for s, b, f in zip(source2, bytesio2, formats2):
            s.save(b, format=f)
        self.dataset = IndexableDataset(
            OrderedDict([('source1', [b.getvalue() for b in bytesio1]),
                         ('source2', [b.getvalue() for b in bytesio2])]),
            axis_labels={'source1': ('batch', 'bytes'),
                         'source2': ('batch', 'bytes')})
        self.common_setup()

    def test_images_from_bytes_example_stream(self):
        stream = ImagesFromBytes(self.example_stream,
                                 which_sources=('source1', 'source2'),
                                 color_mode=None)
        s1, s2 = list(zip(*list(stream.get_epoch_iterator())))
        s1_shape = set(s.shape for s in s1)
        s2_shape = set(s.shape for s in s2)
        actual_s1 = set(reorder_axes(s) for s in self.shapes[:3])
        actual_s2 = set(reorder_axes(s) for s in self.shapes[3:])
        assert actual_s1 == s1_shape
        assert actual_s2 == s2_shape

    def test_images_from_bytes_batch_stream(self):
        stream = ImagesFromBytes(self.batch_stream,
                                 which_sources=('source1', 'source2'),
                                 color_mode=None)
        s1, s2 = list(zip(*list(stream.get_epoch_iterator())))
        s1 = sum(s1, [])
        s2 = sum(s2, [])
        s1_shape = set(s.shape for s in s1)
        s2_shape = set(s.shape for s in s2)
        actual_s1 = set(reorder_axes(s) for s in self.shapes[:3])
        actual_s2 = set(reorder_axes(s) for s in self.shapes[3:])
        assert actual_s1 == s1_shape
        assert actual_s2 == s2_shape

    def test_images_from_bytes_example_stream_convert_rgb(self):
        stream = ImagesFromBytes(self.example_stream,
                                 which_sources=('source1'),
                                 color_mode='RGB')
        s1, s2 = list(zip(*list(stream.get_epoch_iterator())))
        actual_s1_gen = (reorder_axes(s) for s in self.shapes[:3])
        actual_s1 = set((3,) + s[1:] for s in actual_s1_gen)
        s1_shape = set(s.shape for s in s1)
        assert actual_s1 == s1_shape

    def test_images_from_bytes_example_stream_convert_l(self):
        stream = ImagesFromBytes(self.example_stream,
                                 which_sources=('source2'),
                                 color_mode='L')
        s1, s2 = list(zip(*list(stream.get_epoch_iterator())))
        actual_s2_gen = (reorder_axes(s) for s in self.shapes[3:])
        actual_s2 = set((1,) + s[1:] for s in actual_s2_gen)
        s2_shape = set(s.shape for s in s2)
        assert actual_s2 == s2_shape

    def test_axis_labels(self):
        stream = ImagesFromBytes(self.example_stream,
                                 which_sources=('source2',))
        assert stream.axis_labels['source1'] == ('bytes',)
        assert stream.axis_labels['source2'] == ('channel', 'height',
                                                 'width')
        bstream = ImagesFromBytes(self.batch_stream,
                                  which_sources=('source1',))
        assert bstream.axis_labels['source1'] == ('batch', 'channel', 'height',
                                                  'width')
        assert bstream.axis_labels['source2'] == ('batch', 'bytes')

    def test_bytes_type_exception(self):
        stream = ImagesFromBytes(self.example_stream,
                                 which_sources=('source2',))
        assert_raises(TypeError, stream.transform_source_example, 54321,
                      'source2')


class TestMinimumDimensions(ImageTestingMixin):
    def setUp(self):
        rng = numpy.random.RandomState(config.default_seed)
        source1 = []
        source2 = []
        source3 = []
        self.shapes = [(5, 9), (4, 6), (4, 3), (6, 4), (2, 5), (4, 8), (8, 3)]
        for i, shape in enumerate(self.shapes):
            source1.append(rng.normal(size=shape))
            source2.append(rng.normal(size=shape[::-1]))
            source3.append(rng.random_integers(0, 255, size=(3,) + shape)
                           .astype('uint8'))
        self.dataset = IndexableDataset(OrderedDict([('source1', source1),
                                                     ('source2', source2),
                                                     ('source3', source3)]),
                                        axis_labels={'source1':
                                                     ('batch', 'channel',
                                                      'height', 'width'),
                                                     'source3':
                                                     ('batch', 'channel',
                                                      'height', 'width')})
        self.common_setup()

    def test_minimum_dimensions_example_stream(self):
        stream = MinimumImageDimensions(self.example_stream, (4, 5),
                                        which_sources=('source1',
                                                       'source3'))
        it = stream.get_epoch_iterator()
        for example, shp in zip(it, self.shapes):
            assert example[0].shape[0] >= 4 and example[0].shape[1] >= 5
            assert (example[1].shape[1] == shp[0] and
                    example[1].shape[0] == shp[1])
            assert example[2].shape[0] == 3
            assert example[2].shape[1] >= 4 and example[2].shape[2] >= 5

    def test_minimum_dimensions_batch_stream(self):
        stream = MinimumImageDimensions(self.batch_stream, (4, 5),
                                        which_sources=('source1',))
        it = stream.get_epoch_iterator()
        for batch, shapes in zip(it, partition_all(self.batch_size,
                                                   self.shapes)):
            assert (example.shape[0] >= 4 and example.shape[1] >= 5
                    for example in batch[0])
            assert (example.shape[1] == shp[0] and
                    example.shape[0] == shp[1]
                    for example, shp in zip(batch[1], shapes))

    def test_axes_exception(self):
        stream = MinimumImageDimensions(self.example_stream, (4, 5),
                                        which_sources=('source1',))
        assert_raises(NotImplementedError,
                      stream.transform_source_example,
                      numpy.empty((2, 3, 4, 2)),
                      'source1')

    def test_resample_exception(self):
        assert_raises(ValueError,
                      MinimumImageDimensions, self.example_stream, (4, 5),
                      resample='notarealresamplingmode')


class TestFixedSizeRandomCrop(ImageTestingMixin):
    def setUp(self):
        source1 = numpy.zeros((9, 3, 7, 5), dtype='uint8')
        source1[:] = numpy.arange(3 * 7 * 5, dtype='uint8').reshape(3, 7, 5)
        shapes = [(5, 9), (6, 8), (5, 6), (5, 5), (6, 4), (7, 4),
                  (9, 4), (5, 6), (6, 5)]
        source2 = []
        biggest = 0
        num_channels = 2
        for shp in shapes:
            biggest = max(biggest, shp[0] * shp[1] * 2)
            ex = numpy.arange(shp[0] * shp[1] * num_channels).reshape(
                (num_channels,) + shp).astype('uint8')
            source2.append(ex)
        self.source2_biggest = biggest
        axis_labels = {'source1': ('batch', 'channel', 'height', 'width'),
                       'source2': ('batch', 'channel', 'height', 'width')}
        self.dataset = IndexableDataset(OrderedDict([('source1', source1),
                                                     ('source2', source2)]),
                                        axis_labels=axis_labels)
        self.common_setup()

    def test_ndarray_batch_source(self):
        # Make sure that with enough epochs we sample everything.
        stream = RandomFixedSizeCrop(self.batch_stream, (5, 4),
                                     which_sources=('source1',))
        seen_indices = numpy.array([], dtype='uint8')
        for i in range(30):
            for batch in stream.get_epoch_iterator():
                assert batch[0].shape[1:] == (3, 5, 4)
                assert batch[0].shape[0] in (1, 2)
                seen_indices = numpy.union1d(seen_indices, batch[0].flatten())
            if 3 * 7 * 5 == len(seen_indices):
                break
        else:
            assert False

    def test_list_batch_source(self):
        # Make sure that with enough epochs we sample everything.
        stream = RandomFixedSizeCrop(self.batch_stream, (5, 4),
                                     which_sources=('source2',))
        seen_indices = numpy.array([], dtype='uint8')
        for i in range(30):
            for batch in stream.get_epoch_iterator():
                for example in batch[1]:
                    assert example.shape == (2, 5, 4)
                    seen_indices = numpy.union1d(seen_indices,
                                                 example.flatten())
                assert len(batch[1]) in (1, 2)
            if self.source2_biggest == len(seen_indices):
                break
        else:
            assert False

    def test_format_exceptions(self):
        estream = RandomFixedSizeCrop(self.example_stream, (5, 4),
                                      which_sources=('source2',))
        bstream = RandomFixedSizeCrop(self.batch_stream, (5, 4),
                                      which_sources=('source2',))
        assert_raises(ValueError, estream.transform_source_example,
                      numpy.empty((5, 6)), 'source2')
        assert_raises(ValueError, bstream.transform_source_batch,
                      [numpy.empty((7, 6))], 'source2')
        assert_raises(ValueError, bstream.transform_source_batch,
                      [numpy.empty((8, 6))], 'source2')

    def test_window_too_big_exceptions(self):
        stream = RandomFixedSizeCrop(self.example_stream, (5, 4),
                                     which_sources=('source2',))

        assert_raises(ValueError, stream.transform_source_example,
                      numpy.empty((3, 4, 2)), 'source2')

        bstream = RandomFixedSizeCrop(self.batch_stream, (5, 4),
                                      which_sources=('source1',))

        assert_raises(ValueError, bstream.transform_source_batch,
                      numpy.empty((5, 3, 4, 2)), 'source1')


class TestRandom2DRotation(ImageTestingMixin):
    def setUp(self):
        source1 = numpy.zeros((2, 3, 4, 5), dtype='uint8')
        source1[:] = numpy.arange(3 * 4 * 5, dtype='uint8').reshape((3, 4, 5))

        source2 = numpy.empty(2, dtype=object)
        source2[0] = numpy.arange(3 * 4 * 5, dtype='uint8').reshape((3, 4, 5))
        source2[1] = numpy.arange(3 * 4 * 6, dtype='uint8').reshape((3, 4, 6))

        source3 = [source2[0], source2[1]]

        self.source1 = source1
        self.source2 = source2
        self.source3 = source3

        axis_labels = {'source1': ('batch', 'channel', 'height', 'width'),
                       'source2': ('batch', 'channel', 'height', 'width'),
                       'source3': ('batch', 'channel', 'height', 'width')}
        self.dataset = \
            IndexableDataset(OrderedDict([('source1', source1),
                                          ('source2', source2),
                                          ('source3', source3)]),
                             axis_labels=axis_labels)
        self.common_setup()

    def test_format_exceptions(self):
        estream = Random2DRotation(self.example_stream,
                                   which_sources=('source2',))
        bstream = Random2DRotation(self.batch_stream,
                                   which_sources=('source2',))
        assert_raises(ValueError, estream.transform_source_example,
                      numpy.empty((5, 6)), 'source2')
        assert_raises(ValueError, bstream.transform_source_batch,
                      [numpy.empty((7, 6))], 'source2')
        assert_raises(ValueError, bstream.transform_source_batch,
                      [numpy.empty((8, 6))], 'source2')

    def test_maximum_rotation_invalid_exception(self):
        assert_raises(ValueError, Random2DRotation, self.example_stream,
                      maximum_rotation=0.0,
                      which_sources=('source2',))
        assert_raises(ValueError, Random2DRotation, self.example_stream,
                      maximum_rotation=3.1416,
                      which_sources=('source2',))

    def test_invalid_resample_exception(self):
        assert_raises(ValueError, Random2DRotation, self.example_stream,
                      resample='nonexisting')

    def test_random_2D_rotation_example_stream(self):
        maximum_rotation = 0.5
        rng = numpy.random.RandomState(123)
        estream = Random2DRotation(self.example_stream,
                                   maximum_rotation,
                                   rng=rng,
                                   which_sources=('source1',))
        # the C x X x Y image should have equal rotation for all c in C
        out = estream.transform_source_example(self.source1[0], 'source1')
        expected = numpy.array([[[0,  0,  0,  2,  3],
                                 [0,  0,  1,  7,  8],
                                 [0,  5,  6, 12, 13],
                                 [0, 10, 11, 17, 18]],
                                [[0,  0,  0, 22, 23],
                                 [0, 20, 21, 27, 28],
                                 [0, 25, 26, 32, 33],
                                 [0, 30, 31, 37, 38]],
                                [[0,  0,  0, 42, 43],
                                 [0, 40, 41, 47, 48],
                                 [0, 45, 46, 52, 53],
                                 [0, 50, 51, 57, 58]]], dtype='uint8')
        assert_equal(out, expected)

    def test_random_2D_rotation_batch_stream(self):
        rng = numpy.random.RandomState(123)
        bstream = Random2DRotation(self.batch_stream,
                                   maximum_rotation=0.5,
                                   rng=rng,
                                   which_sources=('source1',))
        # each C x X x Y image should have equal rotation for all c in C
        out = bstream.transform_source_batch(self.source1, 'source1')
        expected = numpy.array([[[[0,  0,  0,  2,  3],
                                  [0,  0,  1,  7,  8],
                                  [0,  5,  6, 12, 13],
                                  [0, 10, 11, 17, 18]],
                                 [[0,  0,  0, 22, 23],
                                  [0, 20, 21, 27, 28],
                                  [0, 25, 26, 32, 33],
                                  [0, 30, 31, 37, 38]],
                                 [[0,  0,  0, 42, 43],
                                  [0, 40, 41, 47, 48],
                                  [0, 45, 46, 52, 53],
                                  [0, 50, 51, 57, 58]]],
                                [[[0,  0,  1,  0,  0],
                                  [0,  5,  6,  2,  3],
                                  [0, 10, 11,  7,  8],
                                  [0, 15, 16, 12, 13]],
                                 [[0, 20, 21,  0,  0],
                                  [0, 25, 26, 22, 23],
                                  [0, 30, 31, 27, 28],
                                  [0, 35, 36, 32, 33]],
                                 [[0, 40, 41,  0,  0],
                                  [0, 45, 46, 42, 43],
                                  [0, 50, 51, 47, 48],
                                  [0, 55, 56, 52, 53]]]], dtype='uint8')
        assert_equal(out, expected)

        expected = \
            [numpy.array([[[0,  0,  0,  2,   3],
                           [0,  0,  1,  7,   8],
                           [0,  5,  6,  12, 13],
                           [0,  10, 11, 17, 18]],
                          [[0,  0,  0,  22, 23],
                           [0,  20, 21, 27, 28],
                           [0,  25, 26, 32, 33],
                           [0,  30, 31, 37, 38]],
                          [[0,  0,  0,  42, 43],
                           [0,  40, 41, 47, 48],
                           [0,  45, 46, 52, 53],
                           [0,  50, 51, 57, 58]]], dtype='uint8'),
             numpy.array([[[0,  0,  1,  2,  0,   0],
                           [0,  6,  7,  8,  3,   4],
                           [12, 13, 14, 15, 9,  10],
                           [18, 19, 20, 15, 16, 17]],
                          [[0,  24, 25, 26,  0,  0],
                           [0,  30, 31, 32, 27, 28],
                           [36, 37, 38, 39, 33, 34],
                           [42, 43, 44, 39, 40, 41]],
                          [[0,  48, 49, 50,  0,  0],
                           [0,  54, 55, 56, 51, 52],
                           [60, 61, 62, 63, 57, 58],
                           [66, 67, 68, 63, 64, 65]]], dtype='uint8')]

        rng = numpy.random.RandomState(123)
        bstream = Random2DRotation(self.batch_stream,
                                   maximum_rotation=0.5,
                                   rng=rng,
                                   which_sources=('source2',))
        out = bstream.transform_source_batch(self.source2, 'source2')
        assert_equal(out[0], expected[0])
        assert_equal(out[1], expected[1])

        rng = numpy.random.RandomState(123)
        bstream = Random2DRotation(self.batch_stream,
                                   maximum_rotation=0.5,
                                   rng=rng,
                                   which_sources=('source3',))
        out = bstream.transform_source_batch(self.source3, 'source3')
        assert_equal(out[0], expected[0])
        assert_equal(out[1], expected[1])



# flake8: noqa
from fuel.datasets.base import (Dataset, IterableDataset,
                                IndexableDataset)

from fuel.datasets.hdf5 import H5PYDataset
from fuel.datasets.adult import Adult
from fuel.datasets.binarized_mnist import BinarizedMNIST
from fuel.datasets.celeba import CelebA
from fuel.datasets.cifar10 import CIFAR10
from fuel.datasets.cifar100 import CIFAR100
from fuel.datasets.caltech101_silhouettes import CalTech101Silhouettes
from fuel.datasets.dogs_vs_cats import DogsVsCats
from fuel.datasets.iris import Iris
from fuel.datasets.mnist import MNIST
from fuel.datasets.svhn import SVHN
from fuel.datasets.text import TextFile
from fuel.datasets.billion import OneBillionWord



from fuel.datasets import H5PYDataset
from fuel.utils import find_in_data_path


class Adult(H5PYDataset):
    filename = 'adult.hdf5'

    def __init__(self, which_sets, **kwargs):
        kwargs.setdefault('load_in_memory', True)
        super(Adult, self).__init__(
            file_or_path=find_in_data_path(self.filename),
            which_sets=which_sets, **kwargs
        )



from fuel.converters.ilsvrc2010 import IMAGE_TARS
from fuel.downloaders.base import default_downloader


def fill_subparser(subparser):
    """Sets up a subparser to download the ILSVRC2010 dataset files.

    Note that you will need to use `--url-prefix` to download the
    non-public files (namely, the TARs of images). This is a single
    prefix that is common to all distributed files, which you can
    obtain by registering at the ImageNet website [DOWNLOAD].

    Note that these files are quite large and you may be better off
    simply downloading them separately and running ``fuel-convert``.

    .. [DOWNLOAD] http://www.image-net.org/download-images


    Parameters
    ----------
    subparser : :class:`argparse.ArgumentParser`
        Subparser handling the `ilsvrc2010` command.

    """
    urls = [
        ('http://www.image-net.org/challenges/LSVRC/2010/'
         'ILSVRC2010_test_ground_truth.txt'),
        ('http://www.image-net.org/challenges/LSVRC/2010/'
         'download/ILSVRC2010_devkit-1.0.tar.gz'),
    ] + ([None] * len(IMAGE_TARS))
    filenames = [None, None] + list(IMAGE_TARS)
    subparser.set_defaults(urls=urls, filenames=filenames)
    subparser.add_argument('-P', '--url-prefix', type=str, default=None,
                           help="URL prefix to prepend to the filenames of "
                                "non-public files, in order to download them. "
                                "Be sure to include the trailing slash.")
    return default_downloader



from fuel.downloaders.base import default_downloader


def fill_subparser(subparser):
    """Sets up a subparser to download the MNIST dataset files.

    The following MNIST dataset files are downloaded from Yann LeCun's
    website [LECUN]:
    `train-images-idx3-ubyte.gz`, `train-labels-idx1-ubyte.gz`,
    `t10k-images-idx3-ubyte.gz`, `t10k-labels-idx1-ubyte.gz`.

    Parameters
    ----------
    subparser : :class:`argparse.ArgumentParser`
        Subparser handling the `mnist` command.

    """
    filenames = ['train-images-idx3-ubyte.gz', 'train-labels-idx1-ubyte.gz',
                 't10k-images-idx3-ubyte.gz', 't10k-labels-idx1-ubyte.gz']
    urls = ['http://yann.lecun.com/exdb/mnist/' + f for f in filenames]
    subparser.set_defaults(urls=urls, filenames=filenames)
    return default_downloader



from fuel.datasets import H5PYDataset
from fuel.transformers.defaults import uint8_pixels_to_floatX
from fuel.utils import find_in_data_path


class CIFAR10(H5PYDataset):
    """The CIFAR10 dataset of natural images.

    This dataset is a labeled subset of the ``80 million tiny images``
    dataset [TINY]. It consists of 60,000 32 x 32 colour images in 10
    classes, with 6,000 images per class. There are 50,000 training
    images and 10,000 test images [CIFAR10].

    .. [CIFAR10] Alex Krizhevsky, *Learning Multiple Layers of Features
       from Tiny Images*, technical report, 2009.

    Parameters
    ----------
    which_sets : tuple of str
        Which split to load. Valid values are 'train' and 'test',
        corresponding to the training set (50,000 examples) and the test
        set (10,000 examples). Note that CIFAR10 does not have a
        validation set; usually you will create your own
        training/validation split using the `subset` argument.

    """
    filename = 'cifar10.hdf5'
    default_transformers = uint8_pixels_to_floatX(('features',))

    def __init__(self, which_sets, **kwargs):
        kwargs.setdefault('load_in_memory', True)
        super(CIFAR10, self).__init__(
            file_or_path=find_in_data_path(self.filename),
            which_sets=which_sets, **kwargs)



import os
import re
import sys
from contextlib import contextmanager

import requests
from progressbar import (ProgressBar, Percentage, Bar, ETA, FileTransferSpeed,
                         Timer, UnknownLength)
from six.moves import zip, urllib
from ..exceptions import NeedURLPrefix


@contextmanager
def progress_bar(name, maxval):
    """Manages a progress bar for a download.

    Parameters
    ----------
    name : str
        Name of the downloaded file.
    maxval : int
        Total size of the download, in bytes.

    """
    if maxval is not UnknownLength:
        widgets = ['{}: '.format(name), Percentage(), ' ',
                   Bar(marker='=', left='[', right=']'), ' ', ETA(), ' ',
                   FileTransferSpeed()]
    else:
        widgets = ['{}: '.format(name), ' ', Timer(), ' ', FileTransferSpeed()]
    bar = ProgressBar(widgets=widgets, max_value=maxval, fd=sys.stdout).start()
    try:
        yield bar
    finally:
        bar.update(maxval)
        bar.finish()


def filename_from_url(url, path=None):
    """Parses a URL to determine a file name.

    Parameters
    ----------
    url : str
        URL to parse.

    """
    r = requests.get(url, stream=True)
    if 'Content-Disposition' in r.headers:
        filename = re.findall(r'filename=([^;]+)',
                              r.headers['Content-Disposition'])[0].strip('"\"')
    else:
        filename = os.path.basename(urllib.parse.urlparse(url).path)
    return filename


def download(url, file_handle, chunk_size=1024):
    """Downloads a given URL to a specific file.

    Parameters
    ----------
    url : str
        URL to download.
    file_handle : file
        Where to save the downloaded URL.

    """
    r = requests.get(url, stream=True)
    total_length = r.headers.get('content-length')
    if total_length is None:
        maxval = UnknownLength
    else:
        maxval = int(total_length)
    name = file_handle.name
    with progress_bar(name=name, maxval=maxval) as bar:
        for i, chunk in enumerate(r.iter_content(chunk_size)):
            if total_length:
                bar.update(i * chunk_size)
            file_handle.write(chunk)


def ensure_directory_exists(directory):
    """Create directory (with parents) if does not exist, raise on failure.

    Parameters
    ----------
    directory : str
        The directory to create

    """
    if os.path.isdir(directory):
        return
    os.makedirs(directory)


def default_downloader(directory, urls, filenames, url_prefix=None,
                       clear=False):
    """Downloads or clears files from URLs and filenames.

    Parameters
    ----------
    directory : str
        The directory in which downloaded files are saved.
    urls : list
        A list of URLs to download.
    filenames : list
        A list of file names for the corresponding URLs.
    url_prefix : str, optional
        If provided, this is prepended to filenames that
        lack a corresponding URL.
    clear : bool, optional
        If `True`, delete the given filenames from the given
        directory rather than download them.

    """
    # Parse file names from URL if not provided
    for i, url in enumerate(urls):
        filename = filenames[i]
        if not filename:
            filename = filename_from_url(url)
        if not filename:
            raise ValueError("no filename available for URL '{}'".format(url))
        filenames[i] = filename
    files = [os.path.join(directory, f) for f in filenames]

    if clear:
        for f in files:
            if os.path.isfile(f):
                os.remove(f)
    else:
        print('Downloading ' + ', '.join(filenames) + '\n')
        ensure_directory_exists(directory)

        for url, f, n in zip(urls, files, filenames):
            if not url:
                if url_prefix is None:
                    raise NeedURLPrefix
                url = url_prefix + n
            with open(f, 'wb') as file_handle:
                download(url, file_handle)



import os

import h5py
import numpy

from fuel.converters.base import fill_hdf5_file


def convert_iris(directory, output_directory, output_filename='iris.hdf5'):
    """Convert the Iris dataset to HDF5.

    Converts the Iris dataset to an HDF5 dataset compatible with
    :class:`fuel.datasets.Iris`. The converted dataset is
    saved as 'iris.hdf5'.
    This method assumes the existence of the file `iris.data`.

    Parameters
    ----------
    directory : str
        Directory in which input files reside.
    output_directory : str
        Directory in which to save the converted dataset.
    output_filename : str, optional
        Name of the saved dataset. Defaults to `None`, in which case a name
        based on `dtype` will be used.

    Returns
    -------
    output_paths : tuple of str
        Single-element tuple containing the path to the converted dataset.

    """
    classes = {b'Iris-setosa': 0, b'Iris-versicolor': 1, b'Iris-virginica': 2}
    data = numpy.loadtxt(
        os.path.join(directory, 'iris.data'),
        converters={4: lambda x: classes[x]},
        delimiter=',')
    features = data[:, :-1].astype('float32')
    targets = data[:, -1].astype('uint8').reshape((-1, 1))
    data = (('all', 'features', features),
            ('all', 'targets', targets))

    output_path = os.path.join(output_directory, output_filename)
    h5file = h5py.File(output_path, mode='w')
    fill_hdf5_file(h5file, data)
    h5file['features'].dims[0].label = 'batch'
    h5file['features'].dims[1].label = 'feature'
    h5file['targets'].dims[0].label = 'batch'
    h5file['targets'].dims[1].label = 'index'

    h5file.flush()
    h5file.close()

    return (output_path,)


def fill_subparser(subparser):
    """Sets up a subparser to convert the Iris dataset file.

    Parameters
    ----------
    subparser : :class:`argparse.ArgumentParser`
        Subparser handling the `iris` command.

    """
    return convert_iris



# -*- coding: utf-8 -*-
from fuel.datasets import H5PYDataset
from fuel.transformers.defaults import uint8_pixels_to_floatX
from fuel.utils import find_in_data_path


class SVHN(H5PYDataset):
    """The Street View House Numbers (SVHN) dataset.

    SVHN [SVHN] is a real-world image dataset for developing machine
    learning and object recognition algorithms with minimal requirement
    on data preprocessing and formatting. It can be seen as similar in
    flavor to MNIST [LBBH] (e.g., the images are of small cropped
    digits), but incorporates an order of magnitude more labeled data
    (over 600,000 digit images) and comes from a significantly harder,
    unsolved, real world problem (recognizing digits and numbers in
    natural scene images). SVHN is obtained from house numbers in
    Google Street View images.

    Parameters
    ----------
    which_format : {1, 2}
        SVHN format 1 contains the full numbers, whereas SVHN format 2
        contains cropped digits.
    which_sets : tuple of str
        Which split to load. Valid values are 'train', 'test' and 'extra',
        corresponding to the training set (73,257 examples), the test
        set (26,032 examples) and the extra set (531,131 examples).
        Note that SVHN does not have a validation set; usually you will
        create your own training/validation split using the `subset`
        argument.

    """
    _filename = 'svhn_format_{}.hdf5'
    default_transformers = uint8_pixels_to_floatX(('features',))

    def __init__(self, which_format, which_sets, **kwargs):
        self.which_format = which_format
        super(SVHN, self).__init__(
            file_or_path=find_in_data_path(self.filename),
            which_sets=which_sets, **kwargs)

    @property
    def filename(self):
        return self._filename.format(self.which_format)



from fuel.downloaders.base import default_downloader


def fill_subparser(subparser):
    """Set up a subparser to download the adult dataset file.

    The Adult dataset file `adult.data` and `adult.test` is downloaded from
    the UCI Machine Learning Repository [UCIADULT].

    .. [UCIADULT] https://archive.ics.uci.edu/ml/datasets/Adult

    Parameters
    ----------
    subparser : :class:`argparse.ArgumentParser`
        Subparser handling the adult command.

    """
    subparser.set_defaults(
        urls=['https://archive.ics.uci.edu/ml/machine-learning-databases/'
              'adult/adult.data',
              'https://archive.ics.uci.edu/ml/machine-learning-databases/'
              'adult/adult.test'],
        filenames=['adult.data', 'adult.test'])
    return default_downloader



import os
import zipfile

import h5py
import numpy
from six.moves import range
from PIL import Image

from fuel.converters.base import check_exists, progress_bar
from fuel.datasets import H5PYDataset

IMAGE_FILE = 'img_align_celeba.zip'
ATTRIBUTES_FILE = 'list_attr_celeba.txt'
DATASET_FILES = [IMAGE_FILE, ATTRIBUTES_FILE]
NUM_EXAMPLES = 202599
TRAIN_STOP = 162770
VALID_STOP = 182637
OUTPUT_FILENAME = 'celeba_aligned_cropped.hdf5'


def _initialize_conversion(directory, output_path, image_shape):
    h5file = h5py.File(output_path, mode='w')
    split_dict = {
        'train': {
            'features': (0, TRAIN_STOP),
            'targets': (0, TRAIN_STOP)},
        'valid': {
            'features': (TRAIN_STOP, VALID_STOP),
            'targets': (TRAIN_STOP, VALID_STOP)},
        'test': {
            'features': (VALID_STOP, NUM_EXAMPLES),
            'targets': (VALID_STOP, NUM_EXAMPLES)}}
    h5file.attrs['split'] = H5PYDataset.create_split_array(split_dict)

    targets_dataset = h5file.create_dataset(
        'targets', (NUM_EXAMPLES, 40), dtype='uint8')
    targets_dataset.dims[0].label = 'batch'
    targets_dataset.dims[1].label = 'target'
    targets_dataset[...] = (
        numpy.loadtxt(os.path.join(directory, ATTRIBUTES_FILE), dtype='int32',
                      skiprows=2, usecols=tuple(range(1, 41))) +
        1) / 2

    features_dataset = h5file.create_dataset(
        'features', (NUM_EXAMPLES, 3) + image_shape, dtype='uint8')
    features_dataset.dims[0].label = 'batch'
    features_dataset.dims[1].label = 'channel'
    features_dataset.dims[2].label = 'height'
    features_dataset.dims[3].label = 'width'

    return h5file


@check_exists(required_files=DATASET_FILES)
def convert_celeba_aligned_cropped(directory, output_directory,
                                   output_filename=OUTPUT_FILENAME):
    """Converts the aligned and cropped CelebA dataset to HDF5.

    Converts the CelebA dataset to an HDF5 dataset compatible with
    :class:`fuel.datasets.CelebA`. The converted dataset is saved as
    'celeba_aligned_cropped.hdf5'.

    It assumes the existence of the following files:

    * `img_align_celeba.zip`
    * `list_attr_celeba.txt`

    Parameters
    ----------
    directory : str
        Directory in which input files reside.
    output_directory : str
        Directory in which to save the converted dataset.
    output_filename : str, optional
        Name of the saved dataset. Defaults to
        'celeba_aligned_cropped.hdf5'.

    Returns
    -------
    output_paths : tuple of str
        Single-element tuple containing the path to the converted
        dataset.

    """
    output_path = os.path.join(output_directory, output_filename)
    h5file = _initialize_conversion(directory, output_path, (218, 178))

    features_dataset = h5file['features']
    image_file_path = os.path.join(directory, IMAGE_FILE)
    with zipfile.ZipFile(image_file_path, 'r') as image_file:
        with progress_bar('images', NUM_EXAMPLES) as bar:
            for i in range(NUM_EXAMPLES):
                image_name = 'img_align_celeba/{:06d}.jpg'.format(i + 1)
                features_dataset[i] = numpy.asarray(
                    Image.open(
                        image_file.open(image_name, 'r'))).transpose(2, 0, 1)
                bar.update(i + 1)

    h5file.flush()
    h5file.close()

    return (output_path,)


@check_exists(required_files=DATASET_FILES)
def convert_celeba_64(directory, output_directory,
                      output_filename='celeba_64.hdf5'):
    """Converts the 64x64 version of the CelebA dataset to HDF5.

    This converter takes the aligned and cropped version of the
    CelebA dataset as input and produces a version that's been resized
    to 78x64 pixels and then center cropped to 64x64 pixels.

    Converts the CelebA dataset to an HDF5 dataset compatible with
    :class:`fuel.datasets.CelebA`. The converted dataset is saved as
    'celeba_64.hdf5'.

    It assumes the existence of the following files:

    * `img_align_celeba.zip`
    * `list_attr_celeba.txt`

    Parameters
    ----------
    directory : str
        Directory in which input files reside.
    output_directory : str
        Directory in which to save the converted dataset.
    output_filename : str, optional
        Name of the saved dataset. Defaults to 'celeba_64.hdf5'.

    Returns
    -------
    output_paths : tuple of str
        Single-element tuple containing the path to the converted dataset.

    """
    output_path = os.path.join(output_directory, output_filename)
    h5file = _initialize_conversion(directory, output_path, (64, 64))

    features_dataset = h5file['features']
    image_file_path = os.path.join(directory, IMAGE_FILE)
    with zipfile.ZipFile(image_file_path, 'r') as image_file:
        with progress_bar('images', NUM_EXAMPLES) as bar:
            for i in range(NUM_EXAMPLES):
                image_name = 'img_align_celeba/{:06d}.jpg'.format(i + 1)
                image = Image.open(
                    image_file.open(image_name, 'r')).resize(
                        (64, 78), Image.ANTIALIAS).crop((0, 7, 64, 64 + 7))
                features_dataset[i] = numpy.asarray(image).transpose(2, 0, 1)
                bar.update(i + 1)

    h5file.flush()
    h5file.close()

    return (output_path,)


def convert_celeba(which_format, directory, output_directory,
                   output_filename=None):
    """Converts the CelebA dataset to HDF5.

    Converts the CelebA dataset to an HDF5 dataset compatible with
    :class:`fuel.datasets.CelebA`. The converted dataset is
    saved as 'celeba_aligned_cropped.hdf5' or 'celeba_64.hdf5',
    depending on the `which_format` argument.

    Parameters
    ----------
    which_format : str
        Either 'aligned_cropped' or '64'. Determines which format
        to convert to.
    directory : str
        Directory in which input files reside.
    output_directory : str
        Directory in which to save the converted dataset.
    output_filename : str, optional
        Name of the saved dataset. Defaults to
        'celeba_aligned_cropped.hdf5' or 'celeba_64.hdf5',
        depending on `which_format`.

    Returns
    -------
    output_paths : tuple of str
        Single-element tuple containing the path to the converted dataset.

    """
    if which_format not in ('aligned_cropped', '64'):
        raise ValueError("CelebA format needs to be either "
                         "'aligned_cropped' or '64'.")
    if not output_filename:
        output_filename = 'celeba_{}.hdf5'.format(which_format)
    if which_format == 'aligned_cropped':
        return convert_celeba_aligned_cropped(
            directory, output_directory, output_filename)
    else:
        return convert_celeba_64(
            directory, output_directory, output_filename)


def fill_subparser(subparser):
    """Sets up a subparser to convert the CelebA dataset files.

    Parameters
    ----------
    subparser : :class:`argparse.ArgumentParser`
        Subparser handling the `celeba` command.

    """
    subparser.add_argument(
        "which_format", help="which dataset format", type=str,
        choices=('aligned_cropped', '64'))
    return convert_celeba



# -*- coding: utf-8 -*-

import numpy

from collections import OrderedDict

from fuel import config
from fuel.datasets import IndexableDataset


class Spiral(IndexableDataset):
    u"""Toy dataset containing points sampled from spirals on a 2d plane.

    The dataset contains 3 sources:

    * features -- the (x, y) position of the datapoints
    * position -- the relative position on the spiral arm
    * label -- the class labels (spiral arm)

    .. plot::

        from fuel.datasets.toy import Spiral

        ds = Spiral(classes=3)
        features, position, label = ds.get_data(None, slice(0, 500))

        plt.title("Datapoints drawn from Spiral(classes=3)")
        for l, m in enumerate(['o', '^', 'v']):
            mask = label == l
            plt.scatter(features[mask,0], features[mask,1],
                        c=position[mask], marker=m, label="label==%d"%l)
        plt.xlim(-1.2, 1.2)
        plt.ylim(-1.2, 1.2)
        plt.legend()
        plt.colorbar()
        plt.xlabel("features[:,0]")
        plt.ylabel("features[:,1]")
        plt.show()

    Parameters
    ----------
    num_examples : int
        Number of datapoints to create.
    classes : int
        Number of spiral arms.
    cycles : float
        Number of turns the arms take.
    noise : float
        Add normal distributed noise with standard deviation *noise*.

    """
    def __init__(self, num_examples=1000, classes=1, cycles=1., noise=0.0,
                 **kwargs):
        seed = kwargs.pop('seed', config.default_seed)
        rng = numpy.random.RandomState(seed)
        # Create dataset
        pos = rng.uniform(size=num_examples, low=0, high=cycles)
        label = rng.randint(size=num_examples, low=0, high=classes)
        radius = (2 * pos + 1) / 3.
        phase_offset = label * (2*numpy.pi) / classes

        features = numpy.zeros(shape=(num_examples, 2), dtype='float32')

        features[:, 0] = radius * numpy.sin(2*numpy.pi*pos + phase_offset)
        features[:, 1] = radius * numpy.cos(2*numpy.pi*pos + phase_offset)
        features += noise * rng.normal(size=(num_examples, 2))

        data = OrderedDict([
            ('features', features),
            ('position', pos),
            ('label', label),
        ])

        super(Spiral, self).__init__(data, **kwargs)


class SwissRoll(IndexableDataset):
    """Dataset containing points from a 3-dimensional Swiss roll.

    The dataset contains 2 sources:

    * features -- the x, y and z position of the datapoints
    * position -- radial and z position on the manifold

    .. plot::

        from fuel.datasets.toy import SwissRoll
        import mpl_toolkits.mplot3d.axes3d as p3
        import numpy as np

        ds = SwissRoll()
        features, pos = ds.get_data(None, slice(0, 1000))

        color = pos[:,0]
        color -= color.min()
        color /= color.max()

        fig = plt.figure()
        ax = fig.gca(projection="3d")
        ax.scatter(features[:,0], features[:,1], features[:,2],
                   'x', c=color)
        ax.set_xlim(-1, 1)
        ax.set_ylim(-1, 1)
        ax.set_zlim(-1, 1)
        ax.view_init(10., 10.)
        plt.show()

    Parameters
    ----------
    num_examples : int
        Number of datapoints to create.
    noise : float
        Add normal distributed noise with standard deviation *noise*.

    """
    def __init__(self, num_examples=1000, noise=0.0, **kwargs):
        cycles = 1.5
        seed = kwargs.pop('seed', config.default_seed)
        rng = numpy.random.RandomState(seed)
        pos = rng.uniform(size=num_examples, low=0, high=1)
        phi = cycles * numpy.pi * (1 + 2*pos)
        radius = (1 + 2 * pos) / 3

        x = radius * numpy.cos(phi)
        y = radius * numpy.sin(phi)
        z = rng.uniform(size=num_examples, low=-1, high=1)

        features = numpy.zeros(shape=(num_examples, 3), dtype='float32')
        features[:, 0] = x
        features[:, 1] = y
        features[:, 2] = z
        features += noise * rng.normal(size=(num_examples, 3))

        position = numpy.zeros(shape=(num_examples, 2), dtype='float32')
        position[:, 0] = pos
        position[:, 1] = z

        data = OrderedDict([
            ('features', features),
            ('position', position),
        ])

        super(SwissRoll, self).__init__(data, **kwargs)



#!/usr/bin/env python
"""Fuel dataset downloading utility."""
import argparse
import importlib
import os

import fuel
from fuel import downloaders
from fuel.downloaders.base import NeedURLPrefix

url_prefix_message = """
Some files for this dataset do not have a download URL.

Provide a URL prefix with --url-prefix to prepend to the filenames,
e.g. http://path.to/files/
""".strip()


def main(args=None):
    """Entry point for `fuel-download` script.

    This function can also be imported and used from Python.

    Parameters
    ----------
    args : iterable, optional (default: None)
        A list of arguments that will be passed to Fuel's downloading
        utility. If this argument is not specified, `sys.argv[1:]` will
        be used.

    """
    built_in_datasets = dict(downloaders.all_downloaders)
    if fuel.config.extra_downloaders:
        for name in fuel.config.extra_downloaders:
            extra_datasets = dict(
                importlib.import_module(name).all_downloaders)
            if any(key in built_in_datasets for key in extra_datasets.keys()):
                raise ValueError('extra downloaders conflict in name with '
                                 'built-in downloaders')
            built_in_datasets.update(extra_datasets)
    parser = argparse.ArgumentParser(
        description='Download script for built-in datasets.')
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument(
        "-d", "--directory", help="where to save the downloaded files",
        type=str, default=os.getcwd())
    parent_parser.add_argument(
        "--clear", help="clear the downloaded files", action='store_true')
    subparsers = parser.add_subparsers()
    download_functions = {}
    for name, fill_subparser in built_in_datasets.items():
        subparser = subparsers.add_parser(
            name, parents=[parent_parser],
            help='Download the {} dataset'.format(name))
        # Allows the parser to know which subparser was called.
        subparser.set_defaults(which_=name)
        download_functions[name] = fill_subparser(subparser)
    args = parser.parse_args()
    args_dict = vars(args)
    download_function = download_functions[args_dict.pop('which_')]
    try:
        download_function(**args_dict)
    except NeedURLPrefix:
        parser.error(url_prefix_message)


if __name__ == "__main__":
    main()



from fuel.datasets import H5PYDataset
from fuel.transformers.defaults import uint8_pixels_to_floatX
from fuel.utils import find_in_data_path


class CIFAR100(H5PYDataset):
    """The CIFAR100 dataset of natural images.

    This dataset is a labeled subset of the ``80 million tiny images``
    dataset [TINY]. It consists of 60,000 32 x 32 colour images labelled
    into 100 fine-grained classes and 20 super-classes. There are
    600 images per fine-grained class. There are 50,000 training
    images and 10,000 test images [CIFAR100].

    The dataset contains three sources:
    - features: the images themselves,
    - coarse_labels: the superclasses 1-20,
    - fine_labels: the fine-grained classes 1-100.

    .. [TINY] Antonio Torralba, Rob Fergus and William T. Freeman,
       *80 million tiny images: a large dataset for non-parametric
       object and scene recognition*, Pattern Analysis and Machine
       Intelligence, IEEE Transactions on 30.11 (2008): 1958-1970.

    .. [CIFAR100] Alex Krizhevsky, *Learning Multiple Layers of Features
       from Tiny Images*, technical report, 2009.

    Parameters
    ----------
    which_sets : tuple of str
        Which split to load. Valid values are 'train' and 'test',
        corresponding to the training set (50,000 examples) and the test
        set (10,000 examples). Note that CIFAR100 does not have a
        validation set; usually you will create your own
        training/validation split using the `subset` argument.

    """
    filename = 'cifar100.hdf5'
    default_transformers = uint8_pixels_to_floatX(('features',))

    def __init__(self, which_sets, **kwargs):
        kwargs.setdefault('load_in_memory', True)
        super(CIFAR100, self).__init__(
            file_or_path=find_in_data_path(self.filename),
            which_sets=which_sets, **kwargs)



import six


class DataIterator(six.Iterator):
    """An iterator over data, representing a single epoch.

    Parameters
    ----------
    data_stream : :class:`DataStream` or :class:`Transformer`
        The data stream over which to iterate.
    request_iterator : iterator
        An iterator which returns the request to pass to the data stream
        for each step.
    as_dict : bool, optional
        If `True`, return dictionaries mapping source names to data
        from each source. If `False` (default), return tuples in the
        same order as `data_stream.sources`.

    """
    def __init__(self, data_stream, request_iterator=None, as_dict=False):
        self.data_stream = data_stream
        self.request_iterator = request_iterator
        self.as_dict = as_dict

    def __iter__(self):
        return self

    def __next__(self):
        if self.request_iterator is not None:
            data = self.data_stream.get_data(next(self.request_iterator))
        else:
            data = self.data_stream.get_data()
        if self.as_dict:
            return dict(zip(self.data_stream.sources, data))
        else:
            return data
